{=================================================================
 EXPARSE  Copyright 1996..1999 by
 Roman Krejci-RKS, Pujmanove 1552, 140 00 Prague 4, Czech Republic
 http://www.rksolution.cz, EMail INFO@rksolution.cz
==================================================================}

{
 This unit contains types and routines for implementing
 expressions that return dataset field values
}

unit Dsetexpr;
{$I alldemo.inc}

interface
uses
  Windows, SysUtils, Classes, Forms, DB, DBTables,
  PARSEH, EXPROBJ;

Const
{ Only following field types can occur in expression based on dataset fields }
  ExAcceptedFiedTypes = [ftString,
                       ftSmallint,
                       ftInteger,
                       ftWord,
                       ftBoolean,
                       ftFloat,
                       ftCurrency,
                       ftBCD,
                       ftDate,
                       ftTime,
                       ftDateTime,
                       ftAutoInc];

Type
{$IFDEF VER90}
  tBDEDataset = tDataset;
{$ENDIF}
{ tExpression with tbl field. Also changes default string comparison function
  to string comparison based on dataset locale }
  tDatasetExpression = class(tExpression)
  public
    tbl : tBDEDataset;
    constructor Create(aTbl: tBDEDataset);
  end;

Var
  BaseOpRoot: pexProcRec; { Root of Operation Definition Records linked list }


{ Changes spaces to underscores - necessary for creating operation identifier
  from multiword field names }
Function  SpacesToUnderScore(const s: string):string;

{ Creates operation definition records for dataset fields, adds them to ODRlist list
  and links them to the chain Exparse.DLL iterates when parsing an expression }
Procedure RegisterTableOperations(IncludeCalcFields: Boolean; // expression can contain calculated fields
                                  ODRlist: tStringList;    // List holding identifiers and ODRs
                                  DataSet: TDataSet;       // Dataset in question
                                  var tabOpDescr: string;  // Operations class description
                                  var tabRoot: pexProcRec);// in-> Old root of linked list, out <- new root of linked list

{ Clears list of Operation definition records previously filled with RegisterTableOperations }
Procedure ClearODRList(ODRList: tStringList);

implementation



{******** tDatasetExpression *********}

{ String comparison based on table locale }
Function  TableCmpStrings(Exprhandle: tExprhandle;  Const S1,S2: string): Longint; Stdcall;
var
  l1,l2,l: integer;
begin
  if S1=S2 then Result := 0 else
  if S1='' then Result := -1 else
  if S2='' then result := 1 else begin
    l1 := length(s1);
    l2 := length(s2);
    if l1 < l2 then l := l1 else l := l2;
    With tDatasetExpression(exprGetUserData(Exprhandle)) do begin
      if (tbl<>NIL) and (tbl.Locale<>NIL) then begin
{ Delphi bug - NativeCompareStr typecasts smallint to integer, so we have to
               typecast it back  }
        Result := smallint(NativeCompareStr(tbl.Locale,S1, S2, l));
        if (Result=0) and (l1<>l2) then begin
          if l1 < l2 then result := -1
          else result := 1
        end;
      end else Result := AnsiCompareStr(s1,s2);
    end
  end;
end;

constructor tDatasetExpression.Create(aTbl: tBDEdataset);
begin
  inherited Create;
  Tbl := aTbl;
  if Tbl <> NIL then exprSetCompareTextFunc(handle,@TableCmpStrings);
end;

{********************************************}

Function SpacesToUnderScore(const s: string):string;
var i: integer;
begin
  Result := s;
  For i := 1 to length(result) do
    if result[i]=' ' then result[i] := '_';
end;

{ FldValue function returns value of given field as variant.
  Special attention is paid to lookup fields. Thanks to
  use of As(String,Integer etc.) methods, NULLs are converted
  to empty strings, zeros etc. }
Function FldValue(Fld: tField; FlddataSet: tdataSet): variant;
begin
  Result := NULL;
  if Fld <> NIL then begin
    If Fld.LookUp then begin
      if (Fld.LookUpdataSet <> NIL) and Fld.LookUpdataSet.Active then with Fld.LookUpdataSet do begin
        Try
          if FldDataset <> NIL then
          Result := Lookup(Fld.LookUpKeyFields,FldDataSet.FindField(Fld.KeyFields).Value,Fld.LookUpResultField);
        Except
        End
      end;
    end else begin
     // if Fld Is tmemoField then with Fld As tmemoField do Result := Text else
      case Fld.dataType of
      ftString: Result := Fld.AsString;
      ftSmallint,
      ftInteger:  Result := Fld.AsFloat;
      ftWord,
      ftFloat,
      ftCurrency,
      ftAutoInc,
      ftBCD: Result := Fld.AsFloat;
      ftBoolean: result := Fld.AsBoolean;
      ftDate,
      ftTime,
      ftDateTime: Result := Fld.AsdateTime;
      end;
    end;
  end;
end;

{ This is the actual evaluation function associated with operations returning field value }
Function datasetExprResult(p:pexProcRec; Exprhandle: tExprhandle; Reserved1: Pointer): variant; Stdcall;
begin
  With tDatasetExpression(exprGetUserData(Exprhandle)) do
    Result := FldValue(tbl.FindField(String(P.userLong)),tbl);
end;

Procedure RegisterTableOperations(IncludeCalcFields: Boolean; // expression can contain calculated fields
                                  ODRlist: tStringList;    // List holding identifiers and ODRs
                                  DataSet: TDataSet;       // Dataset in question
                                  var tabOpDescr: string;  // Operations class description
                                  var tabRoot: pexProcRec);// in-> Old root of linked list, out <- new root of linked list
VAR
  S:String;
  p: pexProcRec;
  i,j: integer;
  Fld: tField;
begin
  if ODRList = NIL then exit;
  tabOpDescr := 'Fields of '+Dataset.name+' dataset'#0;
  j := 0;
  with DataSet do For i := 0 to FieldCount-1 do begin
    Fld := Fields[i];
    if (IncludeCalcFields or (not Fld.Calculated))
      and
      (Fld.dataType in ExAcceptedFiedTypes)
    then begin
      S := ANSIUPPERCASE(SpacesToUnderScore(Fld.Fieldname))+#0; // operation identifier in expression
      New(p);
      FillChar(p^,SizeOf(p^),#0);
      with p^ do begin
        ClassIdentif := @tabOpDescr[1];
        case Fld.dataType of
        ftBoolean  : aResult := ext_Bool;
        ftDate,
        ftDateTime : aResult := ext_Date;
        ftAutoInc,
        ftBCD,
        ftCurrency,
        ftFloat,
        ftWord,
        ftInteger,
        ftSmallint : aResult := ext_Number;
        else         aResult := ext_String;
        end;
        ProcAddr  := @datasetExprResult;
{ this will identify the field - workaround for multiword field names: }
        String(UserLong) := Fld.Fieldname;
      end;
      ODRlist.AddObject(s,tObject(p));
      p^.Pidentif := @ODRlist.Strings[j][1];
      NEXTRegisterODR(TabRoot,p); // Links p^ into chain Exparse.dll iterates when parsing an expression.
      inc(j);
    end;
  end;
end;

{ Clears list of Operation definition records previously filled with
  RegisterTableOperations }
Procedure ClearODRList(ODRList: tStringList);
var i: integer;
begin
  if ODRList <> NIL then
  with ODRList do begin
    for i :=0 to count-1 do begin
      String(pexprocrec(objects[i]).UserLong) := '';
      dispose(pexprocrec(objects[i]));
    end;
    Clear;
  end;
end;



end.





