{=================================================================
 EXPARSE  Copyright 1996..1999 by
 Roman Krejci-RKS, Pujmanove 1552, 140 00 Prague 4, Czech Republic
 http://www.rksolution.cz, EMail INFO@rksolution.cz
==================================================================}

{
  This unit implements class (tExpression) that encapsulates
  most of the Exparse.DLL functionality
}

UNIT Exprobj;
{$I alldemo.inc}

interface
USES Classes,SysUtils,
     PARSEH;

Type

{ Exceptions during Expression's compile time }
  ExpressionError = Class(Exception)
                    end;


  TExpression=class
             Private
                Function  GetResultType : texType;
                Function  GetAsString   : exString;
                Function  GetAsNumber   : exNumber;
                Function  GetAsDate     : exDate;
                Function  GetAsBoolean  : exBool;
                Function  GetAsVariant  : variant;
                Function  GetExecErrCode: Integer;
                Procedure SetExecErrCode(I: Integer);
             public
                Handle : tExprHandle;
                constructor Create;
                destructor  Destroy; Override;

                Procedure SetText(Const S: String; ListRoot: pexProcRec);
                Function  GetText:String;
                Function  GetErrText:String;
                procedure PrepareRecalc; { Prepares for fresh recalculation by clearing error flags }
                function  Execute(var ErrCode: longint): Variant;
                function  CheckExecute: Variant;
                  { CheckExecute calls PrepareRecalc.
                    Then performs execute and raises exception if any error occurred }
                Function  HasError:Boolean;
                Function  PeekResult(Var ErrCode:Integer): Variant;
                Function  DependOnODR(ODR: pexProcRec): Boolean;
                  { DependOnODR checks whether expression depends on operation defined by ODR }


                property  ResultType    : texType       Read GetResultType;
                Property  AsString      : exString      Read GetAsString;
                Property  AsNumber      : exNumber      Read GetAsNumber;
                Property  AsDate        : exDate        Read GetAsDate;
                Property  AsBoolean     : exBool        Read GetAsBoolean;
                Property  AsVariant     : variant       Read GetAsVariant;
                Property  ExecErrorCode : Integer       Read GetExecErrCode Write SetExecErrCode;
             end;


Implementation

{--- tExpression -----}

constructor TExpression.Create;
begin  exprGetExprhandle(Handle,Self)
end;
destructor  TExpression.Destroy;
begin  exprFreeExprhandle(Handle)
end;
Function    TExpression.GetText:String;
begin  Result := exprGetText(Handle)
end;
Function    TExpression.GetErrText:String;
begin  Result := exprGetErrText(Handle)
end;
Procedure   TExpression.SetText(Const S: String; ListRoot: pexProcRec);
begin
  case exprSetText(Handle,S,ListRoot) of
  eex_OK                : ;
  eex_MissingOperand    : raise ExpressionError.Create('Missing operand:'#13+QuotedStr(GetErrText));
  eex_UnknownOperation  : raise ExpressionError.Create('Unknown operation:'#13+QuotedStr(GetErrText));
  eex_ArityError        : raise ExpressionError.Create('Incorrect number of operands:'#13+QuotedStr(GetErrText));
  eex_TypeMismatch      : raise ExpressionError.Create('Type mismatch:'#13+QuotedStr(GetErrText));
  eex_MissingLeftPar    : raise ExpressionError.Create('"(" expected:'#13+QuotedStr(GetErrText));
  eex_MissingRightPar   : raise ExpressionError.Create('")" expected:'#13+QuotedStr(GetErrText));
  eex_UnterminatedString: raise ExpressionError.Create('Unterminated string:'#13+QuotedStr(GetErrText));
  eex_TooComplicated    : raise ExpressionError.Create('Expression is too complicated:'#13+QuotedStr(GetErrText));
  end;
end;
Function    TExpression.GetResultType: texType;
begin  Result := exprGetResultType(Handle)
end;
Function    TExpression.GetAsString;
begin
  Result := exprGetAsString(Handle)
end;
Function    TExpression.GetAsNumber;
begin
  Result := exprGetAsNumber(Handle)
end;
Function    TExpression.GetAsDate;
begin
  Result := exprGetAsDate(Handle)
end;
Function    TExpression.GetAsBoolean;
begin
  Result := exprGetAsBoolean(Handle)
end;
Function    TExpression.GetAsVariant;
begin
  Result := exprGetAsVariant(Handle)
end;
Function    TExpression.GetExecErrCode:Integer;
begin Result := exprGetExecErrorCode(Handle)
end;
Procedure    TExpression.SetExecErrCode(I: Integer);
begin exprSetExecErrorCode(Handle,I)
end;
procedure   TExpression.PrepareRecalc;
begin exprPrepareRecalc(Handle);
end;
function    TExpression.Execute(var ErrCode: longint): Variant;
begin
  ErrCode := 0;
  Result := exprExecute(Handle,errcode);
end;
function  TExpression.CheckExecute: Variant;
var ErrCode: longint;
begin
  PrepareRecalc;
  Result := Execute(errcode);
  case Errcode of
  0: ;
  eex_InvalidOpType  : raise Exception.Create('Invalid operand type');
  eex_OverUnderFlow  : raise Exception.Create('Too large/small number');
  eex_DivZero        : raise Exception.Create('Division by zero');
  eex_InvalidOperand : raise Exception.Create('Invalid operand');
  eex_UnknowType     : raise Exception.Create('Unknown result type');
  else raise Exception.Create('Unknown error');
  end;
end;
Function    TExpression.HasError:Boolean;
begin Result := exprGetHasError(Handle)
end;
Function    TExpression.PeekResult(Var ErrCode:Integer): Variant;
begin Result := exprPeekResult(Handle,ErrCode)
end;
Function  TExpression.DependOnODR(ODR: pexProcRec): Boolean;
begin Result := exprContainsOperation (Handle,ODR);
end;

END.


