{=================================================================
 EXPARSE  Copyright 1996..1999 by
 Roman Krejci-RKS, Pujmanove 1552, 140 00 Prague 4, Czech Republic
 http://www.rksolution.cz, EMail INFO@rksolution.cz
==================================================================}

{
  This unit demonstrates the use of Exparse in implementing
  user-defined filters on database table
}

unit Unit1;
{$I ..\alldemo.inc}

interface

uses
  Windows, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,StdCtrls, ExtCtrls,
  DB, DBTables, Grids, DBGrids, IniFiles,
  PARSEH, EXPROBJ, BASEOP, DSETEXPR;

type
  TForm1 = class(TForm)
    DataSource1: TDataSource;
    Table1: TTable;
    DBGrid1: TDBGrid;
    Panel1: TPanel;
    Label1: TLabel;
    LabelCntRecords: TLabel;
    ButtonBuilder: TButton;
    ButtonClose: TButton;
    ButtonApply: TButton;
    ButtonClear: TButton;
    FilterCombo: TComboBox;
    procedure ButtonCloseClick(Sender: TObject);
    procedure ButtonApplyClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ButtonBuilderClick(Sender: TObject);
    procedure ButtonClearClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
    Expr: tExpression;
    TabRoot: pexProcRec;
    tabOpDescr : string;
    ListFldOps : tStringList;
    OldAfterOpen,
    OldBeforeClose: tdatasetNotifyEvent;
    OldFilterRecord: TFilterRecordEvent;
    procedure NewAfterOpen(DataSet: TDataSet);
    procedure NewBeforeClose(DataSet: TDataSet);
    procedure NewFilterRecord(DataSet: TDataSet; var Accept: Boolean);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses ExprBldr;

{$R *.DFM}


{ Form1 }

procedure TForm1.FormCreate(Sender: TObject);
var i: integer;
    ININame,s: string;
begin
  Table1.Close;
  IniName := ChangeFileExt(Paramstr(0),'.INI');
  if FileExists(IniName) then
  With tIniFile.Create(IniName) do Try
    i := 0;
    Repeat
      s := ReadString('Filters',format('H%d',[i]),'');
      inc(i);
      if s <> '' then FilterCombo.Items.Add(s) else break
    Until i > 20;
    s := ReadString('Filters','Current','');
    FilterCombo.Text := s;
  Finally
    Free;
  End;
  OldAfterOpen := Table1.AfterOpen;
  OldBeforeClose := Table1.BeforeClose;
  OldFilterRecord := Table1.OnFilterRecord;
  Table1.AfterOpen := NewAfterOpen;
  Table1.BeforeClose := NewBeforeClose;
  Table1.OnFilterRecord := NewFilterRecord;
  Table1.Open;
end;

procedure TForm1.FormDestroy(Sender: TObject);
var i: Integer;
begin
  Expr.Free;
  Expr := nil;
  ClearODRList(ListFldOps);
  ListFldOps.Free;
  With tIniFile.Create(ChangeFileExt(Paramstr(0),'.INI')) do Try
    EraseSection('Filters');
    for i := 0 to FilterCombo.Items.Count-1 do
      WriteString('Filters',format('H%d',[i]),FilterCombo.Items[i]);
    WriteString('Filters','Current',FilterCombo.Text);
  Finally
    Free;
  End;
end;

procedure TForm1.NewAfterOpen(DataSet: TDataSet);
begin
  if Assigned(OldAfterOpen) then OldAfterOpen(Dataset);
  TabRoot := BaseOpRoot;
  ListFldOps := tStringList.Create;
  RegisterTableOperations(False, {cannot filter on calculated fields}
                          ListFldOps,
                          Dataset,
                          tabOpDescr,
                          TabRoot);
  Expr := tdatasetExpression.Create(table1);
  LabelCntRecords.caption := IntToStr(Dataset.recordcount)+' records - not filtered';
end;

procedure TForm1.NewBeforeClose(DataSet: TDataSet);
begin
  Expr.Free;
  Expr := nil;
  ClearODRList(ListFldOps);
  ListFldOps.Free;
  ListFldOps := nil;
  if Assigned(OldBeforeClose) then OldBeforeClose(Dataset);
end;

procedure TForm1.NewFilterRecord(DataSet: TDataSet;
  var Accept: Boolean);
var errcode: longint;
begin
  if Assigned(OldFilterRecord) then OldFilterRecord(DataSet,Accept);
  if Accept and (Expr <> nil) then begin
    Expr.PrepareRecalc;
    Expr.Execute(errcode);
    if errcode <> 0 then {??} accept := True
    else Accept := Expr.AsBoolean;
  end;
end;


procedure TForm1.ButtonCloseClick(Sender: TObject);
begin
  application.terminate;
end;

procedure TForm1.ButtonApplyClick(Sender: TObject);
begin
  if FilterCombo.Text = '' then table1.filtered := false else begin
    if FilterCombo.Items.IndexOf(FilterCombo.Text) < 0 then
      FilterCombo.Items.Insert(0,FilterCombo.Text);
    Expr.SetText(FilterCombo.Text,TabRoot);
    if Expr.Resulttype <> ext_Bool then begin
      showmessage('Boolean expression expected');
      table1.Filtered := false;
    end else table1.Filtered := True;
  end;
  if table1.Filtered then
    LabelCntRecords.caption := IntToStr(table1.recordcount)+' records (filtered)'
  else
    LabelCntRecords.caption := IntToStr(table1.recordcount)+' records - not filtered';
end;

procedure TForm1.ButtonClearClick(Sender: TObject);
begin
  if FilterCombo.Text <> '' then
  if FilterCombo.Items.IndexOf(FilterCombo.Text) < 0 then
    FilterCombo.Items.Insert(0,FilterCombo.Text);
  FilterCombo.Text := '';
  ButtonApplyClick(Sender);
end;


procedure TForm1.ButtonBuilderClick(Sender: TObject);
var s: string;
begin
  s := FilterCombo.text;
  if _BuildExpr(s,Tabroot,[ext_Bool],@InvokeBuilderHelp) then
    FilterCombo.text := s;
end;



Initialization
  RegisterBaseOpNEXT(BaseOpRoot);


end.
