{=================================================================
 EXPARSE  Copyright 1996..1999 by
 Roman Krejci-RKS, Pujmanove 1552, 140 00 Prague 4, Czech Republic
 http://www.rksolution.cz, EMail INFO@rksolution.cz
==================================================================}

{
  This units contains implementaion of
  engineering and statistical functions
}


UNIT ExMath;
{$I ..\alldemo.inc}


INTERFACE
Uses SysUtils,
     Math,
     Distr;

const
  L2              =0.693147180559945; { natural logarithm of 2 }
  L10             =2.30258509299405;  { natural logarithm of 10 }


    { ***** exponentials  ***** }

function Exp10      (x:Extended):Extended;
function Exp2       (x:Extended):Extended;


{ ***************** STATISTICAL FUNCTIONS *********************** }

function LnGamma    (x:Extended):Extended;
function LnBeta(a,b:Extended):Extended;

    { ***** N(0,1)                                            ***** }

function NormalDensity       (x:Extended):Extended;
function NormalDF            (x:Extended):Extended;
function NormalKvantil       (alpha:Extended):Extended;

    { ***** Chi-square  ***** }

Function ChiSquareDensity    (df:Longint;x:Extended):Extended;
function ChisquareDF         (df:Longint;x:Extended):Extended;
Function ChiSquareKvantil    (df:Longint;alpha:Extended):Extended;

    { ***** Beta  ***** }
Function BetaDensity         (a,b,x:Extended):Extended;
Function BetaDF              (a,b,x:Extended):Extended;
Function BetaKvantil         (a,b,alpha:Extended):Extended;

    { ***** F ***** }

Function Fisher_SnedecorDensity  (df1,df2:Longint;x:Extended):Extended;
function Fisher_SnedecorDF       (df1,df2:Longint;x:Extended):Extended;
function Fisher_SnedecorKvantil  (df1,df2:Longint;alpha:Extended):Extended;

    { ***** Student ***** }

Function StudentDensity      (df:Longint;x:Extended):Extended;
function StudentDF           (df:Longint;x:Extended):Extended;
function StudentKvantil      (df:Longint;alpha:Extended):Extended;

    { ***** Non-parametric **** }

function Binom               (N,M:longint):Extended;
function M_WCFD              (N,M,U:longint):Extended;
function Mann_WhitneyDF      (N,M,U:longint):Extended;
function WilcoxonDF          (N,M,W:longint):Extended;

   { ****** Random values **** }

Function MyRandom(Range:Longint):Extended;
                                            {Range=0   ... R[0,1] continuous}
                                            {Range=N>0 ... R[1..N] discrete}


             { >>>>>>>> } IMPLEMENTATION { <<<<<<<<< }


CONST
  SQRLIMIT     =1E2466;
  EXPLIMIT     =4932*L10;
  NEARZEROLIMIT=2E-4932;
  MaxArrayLen  =6551;

TYPE
  ExtArray= array [0..MaxArrayLen] of Extended;


Procedure InvalidOp;
begin
  raise EInvalidOp.Create('Invalid operand');
end;
Procedure OverFlow;
begin
  raise EOverFlow.Create('Overflow');
end;
Function IsInteger(x: extended): boolean;
begin
  Result := abs(x-round(x)) < NEARZEROLIMIT;
end;

{---------------------------------------------------------------------------}
function Exp10      (x:Extended):Extended;
begin
  Result :=Exp(x*L10);
end;
{---------------------------------------------------------------------------}
function Exp2       (x:Extended):Extended;
begin
  Result:=Exp(x*L2);
end;
{---------------------------------------------------------------------------}
function LnGamma    (x:Extended):Extended;
{ natural logarithm of gamma func. x must be positive integer multiple of 1/2 }
begin
  if (x<0.1) OR not IsInteger(2*x) then InvalidOp;
  Result := LnGammaI(round(2*x));
end;
{ ------------------------------------------------------------------------ }
function LnBeta      (a,b:Extended):Extended;
{ natural logarithm of beta func. both a and b must be positive integer multiples of 1/2 }
Var
  yy,
  xx:Extended;
begin
  if (a<0.1) OR not IsInteger(2*a) then InvalidOp;
  if (b<0.1) OR not IsInteger(2*b) then InvalidOp;
  xx:=LnGamma(a+b);
  yy := LnGamma(a);
  Result:=LnGamma(b) + yy -xx;
end;
{---------------------------------------------------------------------------}
function NormalDensity (x:Extended):Extended;
begin
  if ABS(x)>10 then Result:=0
  else
  Result:=0.398942280401433*exp(-sqr(x)/2)
end;
{---------------------------------------------------------------------------}
function NormalDF (x:Extended):Extended;
{ cumulative distr func of Normal distribution. Error<7.5E-08!  }
begin
  Result := 1 - pNormal(x);
end;
{---------------------------------------------------------------------------}
function NormalKvantil (alpha:Extended):Extended;
{ Inverse of Normal distribution. Error<4.5E-04! }
begin
  if (alpha<=0) OR (alpha>=1) then InvalidOp;
  Result := pNormalInv(1-alpha);
End;
{---------------------------------------------------------------------------}
function ChiSquareDensity(df:Longint;x:Extended):Extended;
var
  dfpul,
  exponent:Extended;
begin
  if df <= 0 then InvalidOp;
  if x<=0 then
  begin
    Result:=0;
    Exit;
  end;
  dfpul:=df/2;
  exponent:=(dfpul-1)*ln(x)-(dfpul*L2+LnGamma(dfpul)+x/2);
  Result:=exp(exponent);
end;
{---------------------------------------------------------------------------}
function ChisquareDF (df:Longint;x:Extended):Extended;
begin
  if df <= 0 then InvalidOp;
  Result := 1 - pChi2(df,x)
end;
{---------------------------------------------------------------------------}
Function ChiSquareKvantil(df:Longint;alpha:Extended):Extended;
begin
  if (alpha<0) OR (alpha>=1) then InvalidOp;
  if df <= 0 then InvalidOp;
  Result := pChi2Inv(df,1-alpha)
end;
{---------------------------------------------------------------------------}
Function BetaDensity(a,b,x:Extended):Extended;
begin
  if (a<0.1) OR not IsInteger(2*a) then InvalidOp;
  if (b<0.1) OR not IsInteger(2*b) then InvalidOp;
  if x<=0 then Result:=0 else
  if x>=1 then Result:=0 else
  Result:=Exp((a-1)*ln(x)+(b-1)*ln(1-x)-LnBeta(a,b));
end;
{---------------------------------------------------------------------------}
function BetaDF(a,b,x:Extended):Extended;
begin
  if (a<0.1) OR not IsInteger(2*a) then InvalidOp;
  if (b<0.1) OR not IsInteger(2*b) then InvalidOp;
  if x<=0 then Result:=0 else
  if x>=1 then Result:=1 else
  Result := pBeta(round(2*a),round(2*b),x)
end;
{ ------------------------------------------------------------- }
Function BetaKvantil(a,b,alpha:Extended):Extended;
begin
  if (alpha<0) OR (alpha>1) then InvalidOp;
  if (a<0.1) OR not IsInteger(2*a) then InvalidOp;
  if (b<0.1) OR not IsInteger(2*b) then InvalidOp;
  Result := pBetaInv(round(2*a),round(2*b), alpha)
end;
{ ------------------------------------------------------------- }
Function Fisher_SnedecorDensity(df1,df2:Longint;x:Extended):Extended;
Var
  exponent:Extended;
begin
  if df1 <= 0 then InvalidOp;
  if df2 <= 0 then InvalidOp;
  if x<=0 then Result:=0 else begin
    exponent:=df1*ln(df1)/2+df2*ln(df2)/2;
    exponent:=exponent-lnbeta(df1/2,df2/2);
    exponent:=exponent+((df1-2)*ln(x)-(df1+df2)*ln(df2+df1*x))/2;
    Result:=exp(Exponent);
  end
end;
{ ------------------------------------------------------------- }
function Fisher_SnedecorDF(df1,df2:Longint;x:Extended):Extended;
begin
  if df1 <= 0 then InvalidOp;
  if df2 <= 0 then InvalidOp;
  if x<=0 then Result:=0 else
    Result:=1-pFDistr(df1,df2,x);
End;
{ ------------------------------------------------------------------- }
function Fisher_SnedecorKvantil(df1,df2:Longint;alpha:Extended):Extended;
begin
  if (alpha<0) OR (alpha>=1) then InvalidOp;
  if df1 <= 0 then InvalidOp;
  if df2 <= 0 then InvalidOp;
  if alpha = 0 then Result := 0 else
    Result := pFdistrInv(df1,df2,1-alpha)
end;
{ ------------------------------------------------------------------- }
Function StudentDensity(df:Longint;x:Extended):Extended;
Var
  exponent:Extended;
begin
  if df <= 0 then InvalidOp;
  exponent:=ln(df)/2+lnbeta(0.5,df/2)+(df+1)*ln(1+x*x/2)/2;
  Result:=exp(-exponent);
end;
{ ------------------------------------------------------------------- }
function StudentDF(df:Longint;x:Extended):Extended;
begin
  if df <= 0 then InvalidOp;
  Result := 1 - pTDistr(df,x)
end;
{ ------------------------------------------------------------------- }
function StudentKvantil(df:Longint;alpha:Extended):Extended;
begin
  if df <= 0 then InvalidOp;
  if (alpha<=0) OR (alpha>=1) then InvalidOp;
  Result := pTDistrInv(df,1-alpha)
end;

{ Non-parametric }

function Binom(N,M:longint):Extended;
{ calculated by using Pascal's triangle }
Var
  MemoryRequired:word;
  C0,C1         :longint;
  EPtr          :^ExtArray;
begin {Binom}
  if (N<0) OR (M<0) OR (M>N)       then InvalidOp
  else
  if M>MaxArrayLen                 then OverFlow;
  if (M=0) OR (M=N) then Result:=1
  else
  if (M=1) OR (M=pred(N)) then Result:=N
  else
  if (M>(N div 2)) then Result:=Binom(N,N-M)
  else
  begin
    MemoryRequired:=succ(M)*SizeOf(Extended);
    GetMem(EPtr,MemoryRequired);
    Try
      FillChar(EPtr^,MemoryRequired,#0);EPtr^[0]:=1;
      For C0:=1 to pred(M) do
      For C1:=C0 downto 1 do
      EPtr^[C1]:=EPtr^[pred(C1)]+EPtr^[C1];
      For C0:=M to N do
      For C1:=M downto 1 do
      EPtr^[C1]:=EPtr^[pred(C1)]+EPtr^[C1];
      Result:=EPtr^[M];
    Finally
      FreeMem(EPtr,MemoryRequired)
    End
  end
end; {Binom}
{ ------------------------------------------------------------------- }
function M_WCFD(N,M,U:longint):Extended;

(***************************************************)
(*                                                 *)
(* Mann-Whitney cumulative frequency distribution- *)
(* Function returns number defined as follows:     *)
(* ----------------------------------------------- *)
(* Let us draw M numbers from the set of N+M       *)
(* natural numbers [1..N+M]. Let W be the sum      *)
(* of M numbers drawn. Now W ranges from M(M+1)/2  *)
(* to MN+M(M+1)/2. Let "u" be the value of W minus *)
(* the lower limit of the above mentioned sum,     *)
(* i.e. u=W-M(M+1)/2. Now u ranges from 0 to MN.   *)
(* Function M_WCFD returns the number of different *)
(* ways the subset of M numbers giving the u value *)
(* less or equal to U can be drawn from the set    *)
(* [1..N+M].                                       *)
(*                                                 *)
(* The value of M_WCFD is calculated by using      *)
(* the recurrence relationship                     *)
(*                                                 *)
(* M_WCFD(N,M,U)=M_WCFD(N-1,M,U-M)+M_WCFD(N,M-1,U) *)
(*                                                 *)
(* with following boundary conditions:             *)
(*                                                 *)
(*      M_WCFD(N,M,-a)=0 for a>0,                  *)
(*      M_WCFD(N,M,0)=1, M_WCFD(N,0,U)=1,          *)
(*      and M_WCFD(N,M,U)=M_WCFD(M,N,U).           *)
(*                                                 *)
(* The matrix needed for calculation is allocated  *)
(* is dynamically and deallocated afterwards.      *)
(*                                                 *)
(***************************************************)
Var
  ArraySize      :longint;
  MemoryRequired :longint;
  EPtr           :^ExtArray;
  C0,C1,C2       :longint;
  C1Limit,C2Limit:longint;

  Function Loc(Row,Col:word):word;
  begin {Loc}
    Loc:=Row*M+pred(Col);
  end;  {Loc}

  Function M_WCDFtemp(a,b,c:longint):Extended;
  begin {M_WCDFtemp}
    if c<0 then             M_WCDFtemp:=0
    else
    if (a=0) OR (b=0) then  M_WCDFtemp:=1
    else
    if a<b then M_WCDFtemp:=M_WCDFtemp(b,a,c)
    else
    if c>(a*b) then         M_WCDFtemp:=M_WCDFtemp(a,b,a*b)
    else                    M_WCDFtemp:=EPtr^[Loc(c,b)];
  end;  {M_WCDFtemp}

Begin {M_WCFD}
  Result := 0;
  if (M<0) OR (N<0)                then InvalidOp
  else
  if (Succ(U)*M)>succ(MaxArrayLen) then OverFlow;

  if U<0 then            Result:=0
  else
  if (M=0) OR (N=0) then Result:=1
  else
  if M>N then            Result:=M_WCFD(M,N,U)
  else
  if U>=M*N then         Result:=Binom(N+M,M)
  else
  if U>(M*N div 2) then  Result:=Binom(N+M,M)-M_WCFD(N,M,M*N-succ(U))
  else
  begin
    ArraySize:=succ(U)*M;
    MemoryRequired:=ArraySize*SizeOf(Extended);
    GetMem(EPtr,MemoryRequired);
    FillChar(EPtr^,MemoryRequired,#0);
    For C0:=1 to N do
    begin
      if C0<M then C1Limit:=C0
      else         C1Limit:=M;
      For C1:=1 to C1Limit do
      begin
        if U>(C0*C1) then C2Limit:=C0*C1
        else              C2Limit:=U;
        For C2:=C2Limit downto 0 do
        if (C0=N) AND (C1=M) AND (C2=U) then
        begin
          Result:=M_WCDFtemp(C0,C1-1,C2)+M_WCDFtemp(C0-1,C1,C2-C1);
          FreeMem(EPtr,MemoryRequired);
          Exit;
        end
        else
          EPtr^[Loc(C2,C1)]:=
          M_WCDFtemp(C0,C1-1,C2)+M_WCDFtemp(C0-1,C1,C2-C1)
      end
    end
  end
end;  {M_WCFD}
{ ------------------------------------------------------------------- }
function Mann_WhitneyDF(N,M,U:longint):Extended;
Begin
  Result:=M_WCFD(N,M,U)/Binom(N+M,M);
End;
{ ------------------------------------------------------------------- }
function WilcoxonDF(N,M,W:longint):Extended;
Begin
  Result:=Mann_WhitneyDF(N,M,W-(M*(M+1) div 2));
End;
{ ------------------------------------------------------------------- }
Function MyRandom(Range:Longint):Extended;
begin
  if (Range>65535) OR
     (Range<0) then InvalidOp;
  if Range=0 then Result:=Random
  else Result:=1+Random(Range);
end;

END.


