{=================================================================
 EXPARSE  Copyright 1996..1999 by
 Roman Krejci-RKS, Pujmanove 1552, 140 00 Prague 4, Czech Republic
 http://www.rksolution.cz, EMail INFO@rksolution.cz
==================================================================}

{
  This unit demonstrates the use of Exparse in implementing
  user-defined filters and calculated fields on database table.
  Please note that now Operation Definition Records are NOT
  linked to the application and are retrieved from
  ODRSbase Ole In Proc server instead.
}

unit Unit1;
{$I ..\alldemo.inc}

interface

uses
  Windows, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,StdCtrls, ExtCtrls,
  Mask, DBCtrls, Menus, DB, DBTables, Grids, DBGrids,
  PARSEH, EXPROBJ, DSETEXPR,
  OleSrvH;

type
  TForm1 = class(TForm)
    DataSource1: TDataSource;
    Table1: TTable;
    PopupMenu1: TPopupMenu;
    Addcalculatedfield1: TMenuItem;
    Editcalculatedfield1: TMenuItem;
    Deletecalculatedfield1: TMenuItem;
    Panel1: TPanel;
    DBGrid1: TDBGrid;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    N1: TMenuItem;
    Close1: TMenuItem;
    Calculatedfield1: TMenuItem;
    cfld_Add: TMenuItem;
    cfld_Edit: TMenuItem;
    cfld_Delete: TMenuItem;
    Filter1: TMenuItem;
    fltr_Edit: TMenuItem;
    fltr_Clear: TMenuItem;
    LabelCntRecords: TLabel;
    Showruntimelicensekey1: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Addcalculatedfield1Click(Sender: TObject);
    procedure DBGrid1DrawColumnCell(Sender: TObject; const Rect: TRect;
      DataCol: Integer; Column: TColumn; State: TGridDrawState);
    procedure Deletecalculatedfield1Click(Sender: TObject);
    procedure Editcalculatedfield1Click(Sender: TObject);
    procedure Close1Click(Sender: TObject);
    procedure fltr_EditClick(Sender: TObject);
    procedure fltr_ClearClick(Sender: TObject);
    procedure Table1AfterPost(DataSet: TDataSet);
    procedure Showruntimelicensekey1Click(Sender: TObject);
  private
    { Private declarations }
    currFltr: String;
    fltrExpr: tExpression;
    fltrTabRoot,
    clcTabRoot: pexProcRec;
    clcTabOpDescr,
    fltrtabOpDescr : string;
    ListCalcFields,  { string:Field name, Object: tExpression}
    clcListFldOps,
    fltrListFldOps : tStringList;
    OldOnCalcFields,
    OldAfterOpen,
    OldBeforeClose: tdatasetNotifyEvent;
    OldFilterRecord: TFilterRecordEvent;
    procedure NewOnCalcFields(DataSet: TDataSet);
    procedure NewAfterOpen(DataSet: TDataSet);
    procedure NewBeforeClose(DataSet: TDataSet);
    procedure NewFilterRecord(DataSet: TDataSet; var Accept: Boolean);
    function  AddAField(Const aName,aCaption,aDispfmt,aExpression: string; aSize: byte): tField;
    procedure DeleteAField(Const aName: string);
    procedure SetFilter(const FilterText: String);
    Procedure ReadINIInfo;
    Procedure StoreINIInfo;
    procedure Updatelabels;
  public
    { Public declarations }
    procedure Idle(Sender: TObject; var Done: Boolean);
  end;

var
  Form1: TForm1;

implementation

uses TypInfo,INIFiles,
     ExprBldr,
     Flddef, Fltrdef;

{$R *.DFM}

{helpers}

{ Adds persistent fields. Dataset must be closed }
procedure UpdatePersistentFields(Owner: tComponent; Dset: tDataset);
var i: integer;
begin
  Dset.FieldDefs.Update;
  For i := 0 to Dset.FieldDefs.Count-1 do begin
    if Dset.FindField(Dset.FieldDefs[i].Name) = nil then { persistent field may already be present }
      Dset.FieldDefs[i].CreateField(Owner);
  end;
end;

{ tForm1 }

procedure TForm1.FormCreate(Sender: TObject);
begin
  clcTabRoot := BaseOpRoot;
  fltrTabRoot := BaseOpRoot;                  { Copy current ODR list root value. This value will
                                                be changed by registration of new operations }
  Table1.Close;                               { Because we shall update persistent fields }
  UpdatePersistentFields(Self,Table1);
  ListCalcFields := tStringList.Create;       { List of added calculated fields }
  clcListFldOps := tStringList.Create;           { List of ODR defining field values }
{ Creating and registering ODR defining field values: }
  RegisterTableOperations(True,               { calculated fields allowed }
                          clcListFldOps,
                          table1,
                          clcTabOpDescr,
                          clcTabRoot);
  OldOnCalcFields := table1.OnCalcFields;     { Assign OnCalcFields event }
  OldAfterOpen := Table1.AfterOpen;
  OldBeforeClose := Table1.BeforeClose;
  OldFilterRecord := Table1.OnFilterRecord;
  Table1.OnCalcFields := NewOnCalcFields;
  Table1.AfterOpen := NewAfterOpen;
  Table1.BeforeClose := NewBeforeClose;
  Table1.OnFilterRecord := NewFilterRecord;
  ReadIniInfo; { Reads calculated field definitions, opens the table and reads grid/form layout }
  Application.OnIdle := Idle;
end;

procedure TForm1.FormDestroy(Sender: TObject);
var i: Integer;
begin
  StoreINIInfo;                   { Stores calculated field definitions and grid/form layout }
  table1.Close;
  clcListFldOps.Free;
  with ListCalcFields do begin    { frees Expression objects for calculated fields }
    for i :=0 to count-1 do Objects[i].Free;
    Free;
  end;
  ListCalcFields := NIL;
end;

procedure TForm1.NewAfterOpen(DataSet: TDataSet);
begin
  if Assigned(OldAfterOpen) then OldAfterOpen(Dataset);
  fltrTabRoot := BaseOpRoot;
  fltrListFldOps := tStringList.Create;
  RegisterTableOperations(False, {cannot filter on calculated fields}
                          fltrListFldOps,
                          Dataset,
                          fltrtabOpDescr,
                          fltrTabRoot);
  fltrExpr := tdatasetExpression.Create(table1);
  Updatelabels;
end;

procedure TForm1.NewBeforeClose(DataSet: TDataSet);
begin
  fltrExpr.Free;
  fltrExpr := nil;
  ClearODRList(fltrListFldOps);
  fltrListFldOps.Free;
  fltrListFldOps := nil;
  if Assigned(OldBeforeClose) then OldBeforeClose(Dataset);
end;

procedure TForm1.NewFilterRecord(DataSet: TDataSet;
  var Accept: Boolean);
var errcode: longint;
begin
  if Assigned(OldFilterRecord) then OldFilterRecord(DataSet,Accept);
  if Accept and (fltrExpr <> nil) then begin
    fltrExpr.PrepareRecalc;
    fltrExpr.Execute(errcode);
    if errcode <> 0 then {??} accept := True
    else Accept := fltrExpr.AsBoolean;
  end;
end;

procedure TForm1.NewOnCalcFields(DataSet: TDataSet);
var ErrCode: Longint;
    Ex: tExpression;
    Fld: tField;
    i: integer;
begin
  if assigned(OldOnCalcFields) then OldOnCalcFields(DataSet);
  if ListCalcFields <> NIL then
  for i :=0 to ListCalcFields.Count-1 do begin {Iterate the list of added calculated fields }
    Fld := Dataset.FindField(ListCalcFields.Strings[i]);
    if Fld=NIL then continue;
    Ex := tExpression(ListCalcFields.Objects[i]); { Objects are Expressions }
    Ex.preparerecalc; { Clears error flags }
    Ex.Execute(errcode); { calculates value }
    if errcode = 0 then begin { No error }
      Fld.value := Ex.AsVariant;
    end else begin
      Fld.value := NULL { any other suggestions ? }
    end
  end;
end;

procedure TForm1.Idle(Sender: TObject; var Done: Boolean);
begin
  { Enable/disable menu items }
  if table1.active then begin
    With DBGrid1 do
     if (SelectedField <> NIL) and (ListCalcFields.IndexOf(SelectedField.FieldName) >= 0)
       then begin
         Editcalculatedfield1.Enabled := True;
         Deletecalculatedfield1.Enabled := True;
         cfld_Edit.Enabled := True;
         cfld_Delete.Enabled := True;
       end else begin
         Editcalculatedfield1.Enabled := false;
         Deletecalculatedfield1.Enabled := false;
         cfld_Edit.Enabled := False;
         cfld_Delete.Enabled := False;
       end;
  end;
end;

procedure TForm1.DBGrid1DrawColumnCell(Sender: TObject; const Rect: TRect;
  DataCol: Integer; Column: TColumn; State: TGridDrawState);
{ User defined calculated fields will have different color }
begin
  if ListCalcFields <> NIL then
  if ListCalcFields.IndexOf(Column.Field.Fieldname) >= 0 then with (Sender as tDBGrid) do begin
    canvas.brush.color := FixedColor;
  end;
  (Sender as tDBGrid).DefaultDrawColumnCell(Rect,DataCol,Column,State);
end;


{ Adding/Editing/Deleting calculated fields }

function TForm1.AddAField(Const aName,aCaption,aDispfmt,aExpression: string; aSize: byte): tField;
var Ex: tdatasetExpression;
    FldClass: tFieldClass;
    Fld: tField;
    PrpInfo : PPropInfo;
begin
  Fld := NIL;
  Ex := tdatasetExpression.Create(table1);
  Try
    Ex.SetText(aExpression,clcTabRoot);
    FldClass := NIL;
    case Ex.ResultType of
    ext_Number: FldClass := tFloatField;
    ext_String: FldClass := tStringField;
    ext_Date  : FldClass := tDateField;
    ext_Bool  : FldClass := tBooleanField;
    end;
    Fld := FldClass.Create(Self);
    Fld.Size := aSize;
    Fld.name := table1.name+aname;
    Fld.calculated := true;
    Fld.Fieldname := aname;
    Fld.displaylabel := acaption;
    if aDispFmt <> '' then begin
      PrpInfo := GetPropInfo(Fld.ClassInfo, 'DisplayFormat');
      if PrpInfo <> nil then { field supports displayformat }
        SetStrProp(Fld, PrpInfo, aDispFmt)
    end;
    Fld.dataset := table1;
    ListCalcFields.AddObject(aname,Ex);
  Except
    Ex.Free;
    Fld.Free;
    raise;
  End;
  Result := Fld;
end;

procedure TForm1.DeleteAField(Const aName: string);
var i: integer;
    Fld: tField;
begin
  i := ListCalcFields.IndexOf(aname);
  if i >= 0 then begin
    ListCalcFields.Objects[i].Free;
    ListCalcFields.Delete(i);
    Fld := table1.FindField(aname);
    if Fld <> NIL then Fld.Free;
  end;
end;


procedure TForm1.Addcalculatedfield1Click(Sender: TObject);
var fname,fcaption,fdispfmt,fexprtext: string;
    aSize: byte;
begin
  fname := '';
  fcaption := '';
  fexprtext := '';
  fdispfmt := '';
  aSize := 0;
  if EditCalcFld(Table1,clcTabRoot,fname,fcaption,fdispfmt,fexprtext,aSize) then begin
    Table1.Close;
    Try
      AddAField(fName,fCaption,fdispfmt,fexprtext,aSize);
    Finally
      Table1.Open;
    End;
  end
end;

procedure TForm1.Deletecalculatedfield1Click(Sender: TObject);
Var aName: string;
begin
  aname := DBGrid1.SelectedField.Fieldname;
  if MessageDlg(Format('Delete calculated field "%s"',[aname]),
                mtConfirmation,[mbYes,mbNo],0)=mrYes then begin
    Table1.Close;
    Try
      DeleteAField(aname);
    Finally
      Table1.Open;
    End;
  end
end;

procedure TForm1.Editcalculatedfield1Click(Sender: TObject);
var OldFname,fname,fcaption,fdispfmt,fexprtext: string;
    aSize: byte;
    Fld: tField;
    i: integer;
    PrpInfo : PPropInfo;
    ColumnIndex: integer;
begin
  ColumnIndex := DBGrid1.SelectedIndex;
  Fld := DBGrid1.SelectedField;
  if Fld <> NIL then begin
    fname := Fld.FieldName;
    i := ListCalcFields.IndexOf(fname);
    if i >= 0 then begin
      OldFname := fName;
      fcaption := Fld.DisplayLabel;
      fexprtext := tExpression(ListCalcFields.Objects[i]).GetText;
      aSize := Fld.Size;
      PrpInfo := GetPropInfo(Fld.ClassInfo, 'DisplayFormat');
      if PrpInfo <> nil then { field supports displayformat }
        fdispfmt := GetStrProp(Fld, PrpInfo)
      else
        fdispfmt := '';
    end else exit;
  end;
  if EditCalcFld(Table1,clcTabRoot,fname,fcaption,fdispfmt,fexprtext,aSize) then begin
    Table1.Close;
    Try
      DeleteAField(OldFname);
      AddAField(fName,fCaption,fdispfmt,fexprtext,aSize);
    Finally
      Table1.Open;
    end;
   { Edited field re-appears as last column in DBGrid, so we restore original column position: }
    DBGrid1.Columns.Items[DBGrid1.Columns.Count-1].Index := ColumnIndex;
  end;
end;

procedure TForm1.Updatelabels;
begin
  if table1.Filtered then begin
    LabelCntRecords.caption := Format('%d records (filter: %s)',[table1.recordcount,CurrFltr]);
  end else begin
    LabelCntRecords.caption :=  Format('%d  records - not filtered',[table1.recordcount]);
  end;
end;

{ Setting filter }
procedure TForm1.SetFilter(const FilterText: String);
begin
  if FilterText = '' then begin
    table1.filtered := false;
    CurrFltr := '';
  end else begin
    fltrExpr.SetText(FilterText,fltrTabRoot);
    table1.Filtered := True;
    CurrFltr := FilterText;
  end;
  Updatelabels;
end;

{ Storing/Reading calculated fields and Form layout }

Procedure TForm1.StoreIniInfo;
Var tabS: String;
    IniF: tIniFile;
    WindowPlacement: tWindowPlacement;

  procedure StoreCalcFields;
  Var I: Integer;
      Fld: tField;
      Ex: tExpression;
      PrpInfo : PPropInfo;
  begin {StoreCalcFields}
    For I := 0 to ListCalcFields.Count-1 do begin
      Ex := tExpression(ListCalcFields.Objects[i]);
      Fld := table1.FindField(ListCalcFields.Strings[i]);
      IniF.WriteString('Calculated','CFLD_NAME'+IntToStr(i),Fld.Fieldname);
      IniF.WriteString('Calculated','CFLD_CAPTION'+IntToStr(i),Fld.Displaylabel);
      PrpInfo := GetPropInfo(Fld.ClassInfo, 'DisplayFormat');
      if PrpInfo <> nil then { field supports displayformat }
         IniF.WriteString('Calculated','CFLD_FMT'+IntToStr(i),GetStrProp(Fld, PrpInfo));
      IniF.WriteInteger('Calculated','CFLD_SIZE'+IntToStr(i),Fld.Size);
      IniF.WriteString('Calculated','CFLD_EXPR'+IntToStr(i),Ex.GetText);
    end;
  end; {StoreCalcFields}

  procedure StoreGridLayout;
  Var I,J: Integer;
  begin {StoreGridLayout}
    For I := 1 to ComponentCount do
    if Components[I-1] Is tDBGrid then With Components[I-1] as tDBGrid do begin
      if (dataSource <> NIL) and (DataSource.dataSet <> NIL) then begin
        tabS := DataSource.dataSet.name
      end else tabS := '';
      if assigned(Columns) then with Columns do
      For j := 0 to Count-1 do With Items[j] do begin
        IniF.WriteInteger('Layout',tabS+'_'+Fieldname+'_Width',Width);
        IniF.WriteInteger('Layout',tabS+'_'+Fieldname+'_Index',J);
      end;
    end;
  end; {StoreGridLayout}

begin {StoreIniInfo}
  IniF:= tIniFile.Create(ChangeFileExt(paramstr(0),'.INI'));
  Try
    IniF.EraseSection('Calculated');
    IniF.EraseSection('Layout');
    StoreCalcFields;
    StoreGridLayout;
    WindowPlacement.length := SizeOf(WindowPlacement);
    GetWindowPlacement(handle,@WindowPlacement);
    IniF.WriteInteger('Layout','State',ord(WindowState));
    with WindowPlacement do begin
      IniF.WriteInteger('Layout','Left', rcNormalPosition.Left);
      IniF.WriteInteger('Layout','Top', rcNormalPosition.Top);
      IniF.WriteInteger('Layout','Right', rcNormalPosition.Right);
      IniF.WriteInteger('Layout','Bottom', rcNormalPosition.Bottom);
    end;
  Finally
    IniF.Free;
  end;
end; {StoreIniInfo}

Procedure TForm1.ReadIniInfo;
Var tabS: String;
    IniF: tIniFile;
    Wstate: Byte;
    WindowPlacement: tWindowPlacement;

  procedure ReadCalcFields;
  Var I: Integer;
      fname,fcaption,fdispfmt,fexprtext: string;
      fSize: integer;
  begin {ReadCalcFields}
    I := 0;
    While true do begin
      fname := IniF.ReadString('Calculated','CFLD_NAME'+IntToStr(i),'');
      if Fname = '' then break;
      fcaption := IniF.ReadString('Calculated','CFLD_CAPTION'+IntToStr(i),'');
      fdispfmt := IniF.ReadString('Calculated','CFLD_FMT'+IntToStr(i),'');
      fexprtext := IniF.ReadString('Calculated','CFLD_EXPR'+IntToStr(i),'');
      fSize := IniF.ReadInteger('Calculated','CFLD_SIZE'+IntToStr(i),0);
      AddAField(fName,fCaption,fdispfmt,fExprtext,fSize);
      inc(i);
    end;
  end; {ReadCalcFields}

  procedure ReadGridLayout;
  Var I,J : integer;
      Coll: tList;
  begin {ReadGridLayout}
    For I := 1 to ComponentCount do
    if Components[I-1] Is tDBGrid then With Components[I-1] as tDBGrid do begin
      if (dataSource <> NIL) and (DataSource.dataSet <> NIL) then begin
        tabs := DataSource.dataSet.name
      end else tabS := '';
      if assigned(Columns) then with Columns do begin
        Coll := tList.Create;
        Try  { read column width and position }
          For j := 0 to Count-1 do Coll.Add(NIL);
          For j := 0 to Count-1 do With Items[j] do begin
            Width := IniF.ReadInteger('Layout',tabS+'_'+Fieldname+'_Width',Width);
            Coll.Items[IniF.ReadInteger('Layout',tabS+'_'+Fieldname+'_Index',J)] := Items[j];
          end;
          For j := 0 to Coll.Count-1 do if Coll.Items[j] <> NIL then tColumn(Coll.Items[j]).Index := j;
        Finally
          Coll.Free;
        end;
      end
    end;
  end; {ReadGridLayout}

begin {ReadIniInfo}
  IniF:= tIniFile.Create(ChangeFileExt(paramstr(0),'.INI'));
  Try
    Try
      ReadCalcFields;
      Table1.Open;
      ReadGridLayout;
      WState := IniF.ReadInteger('Layout','State',ord(wsMaximized));
      FillChar(WindowPlacement,SizeOf(WindowPlacement),#0);
      with WindowPlacement do begin
        length := SizeOf(WindowPlacement);
        case tWindowState(Wstate) of
        wsMaximized: showcmd := SW_SHOWMAXIMIZED;
        else showcmd := SW_SHOWNORMAL;
        end;
        rcNormalPosition.Left  := IniF.ReadInteger('Layout','Left',left);
        rcNormalPosition.Top   := IniF.ReadInteger('Layout','Top',top);
        rcNormalPosition.Right := IniF.ReadInteger('Layout','Right',left + width);
        rcNormalPosition.Bottom:= IniF.ReadInteger('Layout','Bottom', top + height);
      end;
      SetWindowPlacement(handle,@WindowPlacement);
    Except
      ShowMessage('INI file contains invalid information. Delete INI file and rerun the application.');
      Application.Terminate;
    End;
  Finally
    IniF.Free
  End;
end; {ReadIniInfo}

procedure TForm1.Close1Click(Sender: TObject);
begin
  application.terminate
end;

procedure TForm1.fltr_EditClick(Sender: TObject);
Var s: String;
    TestExpr: tExpression;
begin
  s := CurrFltr;
  TestExpr := tdatasetExpression.Create(table1);
  Try
    if GetFilterText(s,fltrTabRoot,TestExpr) then SetFilter(s);
  Finally
    TestExpr.Free;
  End;
end;

procedure TForm1.fltr_ClearClick(Sender: TObject);
begin
  SetFilter('');
end;

procedure TForm1.Table1AfterPost(DataSet: TDataSet);
begin
  UpdateLabels
end;


procedure TForm1.Showruntimelicensekey1Click(Sender: TObject);
var s: string;
begin
  s := GetRuntimeLicKey(CurrentServerClsID);
  InputBox('Runtime license Key','If this box appeared, ODRSBase.LIC file is present.', s);  // No input possible, but we can copy from InputBox.
end;

end.
