{=================================================================
 EXPARSE ver 1.01  Copyright 1996..1999 by
 Roman Krejci-RKS, Pujmanove 1552, 140 00 Prague 4, Czech Republic
 http://www.rksolution.cz, EMail INFO@rksolution.cz
==================================================================}

{
  This unit imports Exparse.DLL types, constants and routines
}


unit PARSEH;
{$I alldemo.inc}


interface
uses windows;

Const
  Stringchar='"';     { encloses string constants in expressions }
  LeftPar   ='(';
  RightPar  =')';
  exMaxParams = 5;    { maximum params in one operation }

{ Character sets }
  Letters      =['@'..'Z','a'..'z','_','$'];
  Numbers      =['0'..'9'];
  IdLetters    =Letters + Numbers;

  SpecialChars =[LeftPar,RightPar,StringChar,' ',';',',','.',':'];
  OtherChars   =[#0..#255]-(IdLetters + SpecialChars);
{ Operation identifier must not contain special chars and must either contain
  ONLY IdLetters or ONLY OtherChars. This means that 'O*R' is not valid identifier,
  whereas both '*' and 'OR' are valid identifiers.  }

{ Error codes }
ErrBase_CompileTime           = 0;
ErrBase_RunTime               = 100;

{A. parse time errors - returned by function exprSetText}
eex_OK                        = 0;
eex_MissingOperand            = ErrBase_CompileTime+1; { missing operand }
eex_UnknownOperation          = ErrBase_CompileTime+2; { unknown operation }
eex_ArityError                = ErrBase_CompileTime+3; { incorrect number of operands }
eex_TypeMismatch              = ErrBase_CompileTime+4; { type mismatch }
eex_MissingLeftPar            = ErrBase_CompileTime+5; { "(" expected }
eex_MissingRightPar           = ErrBase_CompileTime+6; { ")" expected }
eex_UnterminatedString        = ErrBase_CompileTime+7; { unterminated string }
eex_TooComplicated            = ErrBase_CompileTime+8; { too complicated }
{B. Execution time errors. Returned as a var errcode: longint in the  function exprExecute }
eex_UnknownError              = ErrBase_RunTime;   { none of the below }
{eex_CircularRef              = ErrBase_RunTime+1; }
eex_InvalidOpType             = ErrBase_RunTime+2; { Invalid operand type }
eex_OverUnderFlow             = ErrBase_RunTime+3; { numeric Over/Underflow }
eex_DivZero                   = ErrBase_RunTime+4; { division by zero }
eex_InvalidOperand            = ErrBase_RunTime+5; { invalid operand value (i.e. SQRT(-1) }
eex_UnknowType                = ErrBase_RunTime+6; { unknown/unsupported result type (i.e. varUnknown etc) }


{ Attribute of operation - Attrib field of texProcRec record }
atlo_function         = $00; { functions, procedures and variables }
atlo_unary            = $01; { unary operations (unary -,+, logical not) }
atlo_mult             = $02; { multiplicative binary operations ( *, /, AND ) }
atlo_add              = $03; { additive binary operations ( +, -, OR ) }
atlo_relat            = $04; { relational binary operators ( =, <=, <> etc. ) }

Type
{ Parameter and result types }
  texType = (ext_Null,       { No param }
             ext_Number,     { Number }
             ext_String,     { String }
             ext_Date,       { Date }
             ext_Bool);      { Boolean }

  texTypes = Set of texType;

{ actual type of result }  
  exNumber     = Double;
  exString     = AnsiString;
  exDate       = tDateTime;
  exBool       = wordbool;

{ expression objects are manipulated through handles }
  tExprHandle   = tHandle;

{Operation definition record:}
  pexProcRec  =^texProcRec;
  texProcRec  = packed record
             PIdentif : pChar;
            {----------------------------------------------------------------------
             Points to operation identifier (i.e. '<>').
             The identifier must not contain spaces and must all be in uppercase.
            -----------------------------------------------------------------------}
             Attrib: byte; { Precedences of operations }
            {-------------------------------------------------------
             Attrib values:
             0 : functions, procedures and "variables"
             1 : unary operations (unary -,+, logical not)
             2 : multiplicative binary operations ( *, /, AND )
             3 : additive binary operations ( +, -, OR )
             4 : relational binary operators ( =, <=, <> etc. )
             -------------------------------------------------------}
             ProcAddr : pointer;
            {----------------------------------
             Address of the evaluation routine
             ----------------------------------}
             ClassIdentif : pChar;
            {-------------------------------------------------------------------------------
             Points to the "class identifier" of an operation (i.e. 'Arithmetic functions'.
             The class identifier can be used in the Expression builder. Optional.
             -------------------------------------------------------------------------------}
             HelpFilename: pChar;  { Help file name. Optional }
             HelpContext: Longint; { Helpcontext within the help file. Optional }
             aResult,
             Param1,
             Param2,
             Param3,
             param4,
             param5 : texType;
            {---------------------------
             Result and parameter types
             ---------------------------}
             Next    : pexProcRec;
            {-----------------------------------------------------------
             Points to the next registered operation. Managed within
             registration/unregistration routines - do not touch
             -----------------------------------------------------------}
             Counter: Longint;  { Registration counter - do not touch }
             UserPtr: Pointer;  { Never used by exparseDLL except for the proc UserRegisterODR }
             UserLong,          { Never used by exparseDLL }
             IdentifChk: Longint; { Identifier checksum -  do not touch }
           end;

{ Expression callbacks }
 tCompareTextFunc    = function (Exprhandle: tExprhandle;  Const S1,S2: string): Longint; stdcall;
 tStringVariantFunc  = function (Exprhandle: tExprhandle;  Const S: string; VAR V: Variant): wordbool; stdcall;

{ evaluation routines:
  p         - points to the texprocrec being executed
  ExprHandle- handle of the expression being executed
  Reserved  - Must be NIL
}
  TFunc0     = function(p:pexProcRec;ExprHandle:tExprHandle;Reserved: Pointer) : variant; stdcall;
  TFunc1     = function(p:pexProcRec;ExprHandle:tExprHandle;Reserved: Pointer;Const param1: variant):variant; stdcall;
  TFunc2     = function(p:pexProcRec;ExprHandle:tExprHandle;Reserved: Pointer;Const param1,param2: variant): variant; stdcall;
  TFunc3     = function(p:pexProcRec;ExprHandle:tExprHandle;Reserved: Pointer;Const param1,param2,param3: variant): variant;  stdcall;
  TFunc4     = function(p:pexProcRec;ExprHandle:tExprHandle;Reserved: Pointer;Const param1,param2,param3,param4: variant): variant; stdcall;
  TFunc5     = function(p:pexProcRec;ExprHandle:tExprHandle;Reserved: Pointer;Const param1,param2,param3,param4,param5: variant): variant; stdcall;
{ dtto, but now the operation yields no result : }
  TProc0     = procedure(p:pexProcRec;ExprHandle:tExprHandle;Reserved: Pointer); stdcall;
  TProc1     = procedure(p:pexProcRec;ExprHandle:tExprHandle;Reserved: Pointer;Const param1: variant); stdcall;
  TProc2     = procedure(p:pexProcRec;ExprHandle:tExprHandle;Reserved: Pointer;Const param1,param2: variant); stdcall;
  TProc3     = procedure(p:pexProcRec;ExprHandle:tExprHandle;Reserved: Pointer;Const param1,param2,param3: variant); stdcall;
  TProc4     = procedure(p:pexProcRec;ExprHandle:tExprHandle;Reserved: Pointer;Const param1,param2,param3,param4: variant); stdcall;
  TProc5     = procedure(p:pexProcRec;ExprHandle:tExprHandle;Reserved: Pointer;Const param1,param2,param3,param4,param5: variant); stdcall;

{----- DLL import -------}

Procedure exprGetExprhandle     (var H: tExprHandle; UserData: pointer); stdCall;
{ Gets handle for an expression. Must be eventually fred via exprFreeExprhandle }

Procedure exprFreeExprhandle    (var H: tExprHandle); stdCall;
{ Frees handle previously obtained via exprGetExprhandle }

Function  exprGetUserData(H: tExprHandle):pointer; stdCall;
{ Given an expression handle, returns the UserData passed in exprGetExprhandle }

Function  exprSetText           (H: tExprHandle; Const S:String; ListRoot: pexProcRec): longint; stdCall;
{ Parses  the expression's text (i.e. '2+58-sin(2)') . The root of the linked list
  of the operation definition records must be also provided. Returns error code if the
  string cannot be parsed (see error codes) }

Function  exprGetText           (H: tExprHandle):String; stdCall;
{ Gets the last successfully set expression's text }

Function  exprGetErrText        (H: tExprHandle):String; stdCall;
{ Gets the erroneous part of expression string after exprSetText returned an error }

function  exprGetResultType     (H: tExprHandle):texType; stdCall;
{ Gets the expression's result type. Defined immediately after exprSetText, except for spreadsheets }

procedure exprPrepareRecalc     (H: tExprHandle); stdCall;
{ Clears internal error flags and prepares for fresh recalculation }

function  exprExecute           (H: tExprHandle; VAR errcode: longint):Variant; stdCall;
{ executes an expression }

{ procedure exprFinishRecalc      (H: tExprHandle); stdCall;
  obsolete - should not be used. }

function  exprGetAsString       (H: tExprHandle):exString; stdCall;
{ after execute has been performed, returns the result as  string }

function  exprGetAsNumber       (H: tExprHandle):exNumber; stdCall;
{ after execute has been performed, returns the result as  number }

function  exprGetAsDate         (H: tExprHandle):exDate; stdCall;
{ after execute has been performed, returns the result as  datetime }

function  exprGetAsBoolean      (H: tExprHandle):exBool; stdCall;
{ after execute has been performed, returns the result as  boolean }

function  exprGetAsVariant      (H: tExprHandle):variant; stdCall;
{ after execute has been performed, returns the result as  variant }

function  exprGetExecErrorCode  (H: tExprHandle):Integer; stdCall;
{ after execute has been performed, returns the error code }

Procedure exprSetExecErrorCode  (H: tExprHandle;I:Integer); stdCall;
{ you can set the error code during expression's execution }

Function  exprGetHasError       (H: tExprHandle): wordBool; stdCall;
{ shortcut for exprGetExecErrorCode <> 0 }

Function  exprPeekResult        (H: tExprHandle;Var ErrCode:Integer):Variant;stdCall;
{ exprGetExecErrorCode and exprGetAsVariant in one }

Procedure exprSetCompareTextFunc(H: tExprHandle; FuncAddr: Pointer); stdCall;
{ Sets the address of the user defined string comparison function.
  The function whose address is passes must be of type tCompareTextFunc.
  Parameters passed are:
    handle of the expression performing the comparison
    two strings to be compared.

    returns: -1 if s1 < s2, 0 if s1=s2, 1 if s1 > s2

  Default comparison uses AnsiCompareStr
}

Procedure exprSetIsNumberFunc   (H: tExprHandle; FuncAddr: Pointer); stdCall;
{ Changes default Number constant recognition logic to the user defined one.
  The function whose address is passes must be of type tStringVariantFunc.
  Parameters passed are:
    handle of the expression performing the comparison
    string to be parsed

    returns: true if s has been identified as a string representation of a numerical constant
             (in this case it should return the number in V parameter),
             false otherwise
             (in this case it could modify V freely)
}

Procedure exprSetIsStringFunc   (H: tExprHandle; FuncAddr: Pointer); stdCall;
{ Changes default string constant recognition logic to the user defined one.
  The function whose address is passes must be of type tStringVariantFunc.
  Parameters passed are:
    handle of the expression performing the comparison
    string to be parsed

    returns: true if s has been identified as a string constant
             (in this case it should return the string in V parameter),
             false otherwise
             (in this case it could modify V freely)
}

Procedure exprSetIsDateFunc     (H: tExprHandle; FuncAddr: Pointer); stdCall;
{ Changes default Date constant recognition logic to the user defined one.
  The function whose address is passes must be of type tStringVariantFunc.
  Parameters passed are:
    handle of the expression performing the comparison
    string to be parsed

    returns: true if s has been identified as a string representation of a date constant
             (in this case it should return the date in V parameter),
             false otherwise
             (in this case it could modify V freely)
}

Procedure exprSetIsBooleanFunc  (H: tExprHandle; FuncAddr: Pointer); stdCall;
{ Changes default Boolean constant recognition logic to the user defined one.
  The function whose address is passes must be of type tStringVariantFunc.
  Parameters passed are:
    handle of the expression performing the comparison
    string to be parsed

    returns: true if s has been identified as a string representation of a boolean constant
             (in this case it should return corresponding value (true/false)  in V parameter),
             false otherwise
             (in this case it could modify V freely)
}

Function  exprCompareText       (H: tExprHandle;Const S1,S2: string): Integer; stdCall;
{ performs string comparison with comparison logic currently used by Expression with handle H }

Function  exprIsNumber          (H: tExprHandle;Const S: string; VAR number: Variant): wordBool;stdCall;
{ Detects number constants string representation with logic currently used by Expression with handle H }

Function  exprIsString          (H: tExprHandle;Const S: string; VAR str: Variant): wordBool; stdCall;
{ Detects string constants with logic currently used by Expression with handle H }

Function  exprIsdate            (H: tExprHandle;Const S: string; VAR dmy: Variant): wordBool; stdCall;
{ Detects date constants string representation with logic currently used by Expression with handle H }

Function  exprIsBoolean         (H: tExprHandle;Const S: string; VAR bool: Variant): wordBool; stdCall;
{ Detects boolean constants string representation with logic currently used by Expression with handle H }

function  exprIsDefaultNumber   (Const S: string; VAR number: Variant): wordBool; stdCall;
{ Default number constant recognition logic:  Either VAL or StrToFloat suceeds. }

function  exprIsDefaultString   (Const S: string; VAR str: Variant): wordBool; stdCall;
{ Default string constant recognition logic:  Characters are enclosed between "". A '"' character
  can be included in the string as two succesive '"' characters }

function  exprIsDefaultDate     (Const S: string; VAR dmy: Variant): wordBool; stdCall;
{ Default date constant recognition logic: StrTodate succeeds with period as date separator }

function  exprIsDefaultBoolean  (Const S: string; VAR bool: Variant): wordBool; stdCall;
{ Default boolean constant recognition logic:
  S equals to any of the boolean constant string representation
  (Default 'TRUE' and 'FALSE' - can be changed via exprSetBoolStrings
}

Procedure exprSetBoolStrings    (Const STrue,Sfalse: String);stdcall;
{ Sets boolean constants string representation. Default = 'TRUE' and 'FALSE' }

Procedure NextRegisterODR       (var ListRoot:pexProcRec; pProcRec :pexProcRec); stdcall;
{ Links Operation Definition Records through the NEXT member and increments pProcRec.counter
  Records must be linked via NEXT member so the EXPARSE.DLL is able to walk through the chain }

Procedure UserRegisterODR       (var ListRoot:pexProcRec; pProcRec :pexProcRec; VAR total: longint); stdcall;
{ Links Operation Definition Records through the UserPtr member and increments total variable

  List root : in -> old root value, out <- new root value = pProcRec passed (if not NIL)
  pProcRec  : address of the operation definition record to be linked

  The implementation of the two above listed routines follows:

Procedure NextRegisterODR(var ListRoot:pexProcRec; pProcRec :pexProcRec);
var
  p:pexProcRec;
begin
  if pProcRec = NIL then exit;
  if pProcRec.Counter = 0 then begin
    pProcRec.IdentifChk := .. calculates identifier checksum ...
    pProcRec.next := ListRoot;
    ListRoot := pProcRec;
  end;
  Inc(pProcRec.Counter);
end;

Procedure UserRegisterODR(var ListRoot:pexProcRec; pProcRec :pexProcRec; VAR total: longint);
var
  p:pexProcRec;
begin
  if pProcRec = NIL then exit;
  pProcRec.UserPtr := ListRoot;
  ListRoot := pProcRec;
  inc(Total);
end;

}

Function  exprContainsOperation (H: tExprHandle; pProcRec :pexProcRec): wordBool; stdCall;
{ Detects whether expression with handle H depends on Operation Definition Record
  that pProcRec points to. New in version 1.01  }

{--- DLL import END ---}


implementation


const dllname='EXPARSE.DLL';

Procedure exprGetExprhandle     ;external dllname index 1;
Procedure exprFreeExprhandle    ;external dllname index 2;
Function  exprGetUserData       ;external dllname index 3;
Function  exprSetText           ;external dllname index 4;
Function  exprGetText           ;external dllname index 5;
Function  exprGetErrText        ;external dllname index 6;
function  exprGetResultType     ;external dllname index 7;
procedure exprPrepareRecalc     ;external dllname index 8;
function  exprExecute           ;external dllname index 9;
procedure exprFinishRecalc      ;external dllname index 10;
function  exprGetAsString       ;external dllname index 11;
function  exprGetAsNumber       ;external dllname index 12;
function  exprGetAsdate         ;external dllname index 13;
function  exprGetAsBoolean      ;external dllname index 14;
function  exprGetAsVariant      ;external dllname index 15;
function  exprGetExecErrorCode  ;external dllname index 16;
Procedure exprSetExecErrorCode  ;external dllname index 17;
Function  exprGethasError       ;external dllname index 18;
Function  exprPeekResult        ;external dllname index 19;
function  exprIsDefaultNumber   ;external dllname index 20;
Procedure exprSetIsNumberFunc   ;external dllname index 21;
function  exprIsDefaultString   ;external dllname index 22;
Procedure exprSetIsStringFunc   ;external dllname index 23;
function  exprIsDefaultdate     ;external dllname index 24;
Procedure exprSetIsdateFunc     ;external dllname index 25;
Procedure exprSetBoolStrings    ;external dllname index 26;
function  exprIsDefaultBoolean  ;external dllname index 27;
Procedure exprSetIsBooleanFunc  ;external dllname index 28;
Procedure exprSetCompareTextFunc;external dllname index 29;
Function  exprCompareText       ;external dllname index 30;
Function  exprIsNumber          ;external dllname index 31;
Function  exprIsString          ;external dllname index 32;
Function  exprIsdate            ;external dllname index 33;
Function  exprIsBoolean         ;external dllname index 34;
Procedure NextRegisterODR       ;external dllname index 35;
Procedure UserRegisterODR       ;external dllname index 36;
Function  exprContainsOperation ;external dllname index 37; { new in version 1.01 }


end.

